#include <bitmapfonts.c>
#include <gl/shaderprocs.c>
#include <array_1d.c>
#include <rnd/fgetline.c>
#include <rnd/explode.c>
#include <rnd/strconcat.c>
#include <rnd/sq.c>

// Dipole kick: b*(-1,0)
// Quadrupole kick: k*(-x,y)
// Sextupole kick: (1/2)s*(-x*x +y*y, 2x*y)
// Octupole kick: (1/6)o*(-x*x*x +3x*y*y, 3x*x*y -y*y*y)
// Decapole kick: (1/24)d*(-x*x*x*x +6x*x*y*y -y*y*y*y, 4x*x*x*y -4x*y*y*y)
// X[n+1] = x*X[n]+y*Y[n]; Y[n+1] = x*Y[n]-y*X[n]

double factorial(const int n)
{
	if (n<=1) return 1;
	else return n*factorial(n-1);
}

double choose(const int n,const int i)
{
	return factorial(n)/(factorial(i)*factorial(n-i));
}

char *mpterm(const int ax,const int n)
{ // ax=0 gives x component of kick, ax=1 is y component; n=1 for quadrupole, n=2 sextupole etc.
	int i,j; char str[100]; static char ret[1000];
	char *s=ret;
	for (i=ax;i<=n;i+=2)
	{
		if (i%4==0 || i%4==3) *s='-'; else *s='+';
		s++;
		j=choose(n,i);
		if (j>1) {sprintf(str,"%d*",j); strcpy(s,str); s+=strlen(str);}
		for (j=0;j<n;j++) {*s=(j+i>=n?'y':'x'); s++; if (j+1<n) {*s='*'; s++;}}
	}
	*s=0;
	return ret;
}

typedef struct
{
	double x0,x1,xp0,xp1,y0,y1,yp0,yp1, // Range
		xmin,xmax,ymin,ymax; // Collimation
	int baseres,colbase;
	LSet iters,lattice; // LSet(int), LSet(char **)
} Problem;

Problem Problem_new(void)
{
	Problem ret;
	ret.x0=-1; ret.x1=1; ret.xp0=-1; ret.xp1=1;
	ret.y0=-1; ret.y1=1; ret.yp0=-1; ret.yp1=1;
	ret.xmin=-1e20; ret.xmax=1e20; ret.ymin=-1e20; ret.ymax=1e20;
	ret.baseres=32;
	ret.colbase=3;
	ret.iters=LSet(int); ret.lattice=LSet(char **);
	return ret;
}

void Problem_free(Problem p)
{
	LSet_free(p.iters);
	for (int n=p.lattice.m-1;n>=0;n--) explodefree(((char ***)p.lattice.a)[n]);
	LSet_free(p.lattice);
}

Problem Problem_load(const char *file)
{
	Problem ret=Problem_new();
	int n,i; FILE *in=fopen(file,"rb"); if (!in) return ret;
	for (char *s;s=fgetline(in);free(s))
	{
		for (char *t=s;*t;t++) if (*t=='#') {*t=0; break;}
		char **e=explodews(s),used=0;
		if (explodesegs(e)>0)
		{
			//for (n=0;n<explodesegs(e);n++) printf("[%d]'%s' ",n,e[n]); printf("\n");
			if (!strcmpi(e[0],"xrange")) {ret.x0=atof(e[1]); ret.x1=atof(e[2]);}
			else if (!strcmpi(e[0],"xprange")) {ret.xp0=atof(e[1]); ret.xp1=atof(e[2]);}
			else if (!strcmpi(e[0],"yrange")) {ret.y0=atof(e[1]); ret.y1=atof(e[2]);}
			else if (!strcmpi(e[0],"yprange")) {ret.yp0=atof(e[1]); ret.yp1=atof(e[2]);}
			else if (!strcmpi(e[0],"xcollimator")) {ret.xmin=atof(e[1]); ret.xmax=atof(e[2]);}
			else if (!strcmpi(e[0],"ycollimator")) {ret.ymin=atof(e[1]); ret.ymax=atof(e[2]);}
			else if (!strcmpi(e[0],"iterations"))
			{
				for (n=1;n<explodesegs(e);n++) {i=atoi(e[n]); LSet_add(&ret.iters,&i);}
			}
			else if (!strcmpi(e[0],"D") || !strcmpi(e[0],"K"))
			{
				LSet_add(&ret.lattice,&e);
				used=1;
			}
			else if (!strcmpi(e[0],"baseres")) ret.baseres=atoi(e[1]);
		}
		if (!used) explodefree(e);
	}
	fclose(in);
	if (ret.iters.m==0) {n=100; LSet_add(&ret.iters,&n);}
	for (ret.colbase=3;ret.colbase*ret.colbase*ret.colbase<1+ret.iters.m;ret.colbase++);
	return ret;
}

char *Problem_code(const Problem p,const char userecurrence=0)
{ // Constructs a string containing the shader program
	char *ret,str[500];
	streqi(&ret,"uniform float jx,jxp,jy,jyp;\n\
void main()\n\
{\n");
	sprintf(str,"	float x=floor(gl_TexCoord[0].x*%d)/%d,\n\
		y=floor(gl_TexCoord[0].y*%d)/%d,\n\
		xp=floor((gl_TexCoord[0].x-x)*%d)/%d,\n\
		yp=floor((gl_TexCoord[0].y-y)*%d)/%d;\n\
	x=%.9lg%+.9lg*(x+jx/%d); xp=%.9lg%+.9lg*(xp+jxp/%d);\n\
	y=%.9lg%+.9lg*(y+jy/%d); yp=%.9lg%+.9lg*(yp+jyp/%d);\n",
		p.baseres,p.baseres,
		p.baseres,p.baseres,
		p.baseres*p.baseres,p.baseres,
		p.baseres*p.baseres,p.baseres,
		p.x0,p.x1-p.x0,p.baseres,p.xp0,p.xp1-p.xp0,p.baseres,
		p.y0,p.y1-p.y0,p.baseres,p.yp0,p.yp1-p.yp0,p.baseres);
	strconcat(&ret,str);
	if (userecurrence) strconcat(&ret,"	float X,Y,oX;\n");
	int n,i,m,*it=p.iters.a,maxit=it[p.iters.m-1];
	sprintf(str,"	%s i;\n\
	for (i=0;i<=%d",
		maxit>15000000?"int":"float", // int-to-int comparisons are slow on graphics card!  So use single-precision float as loop variable if possible
		maxit);
	strconcat(&ret,str);
	char bothx=(p.xmin>-1e20 && p.xmax<1e20),bothy=(p.ymin>-1e20 && p.ymax<1e20);
	if (!bothy || p.xmin>-1e20) {sprintf(str," && x>=%.9lg",p.xmin); strconcat(&ret,str);}
	if (!bothy || p.xmax<1e20) {sprintf(str," && x<=%.9lg",p.xmax); strconcat(&ret,str);}
	if (!bothx || p.ymin>-1e20) {sprintf(str," && y>=%.9lg",p.ymin); strconcat(&ret,str);}
	if (!bothx || p.ymax<1e20) {sprintf(str," && y<=%.9lg",p.ymax); strconcat(&ret,str);}
	strconcat(&ret,";i++)\n\
	{\n");
	char ***ea=p.lattice.a,**e;
	for (n=0;n<p.lattice.m;n++)
	{
		e=ea[n]; m=explodesegs(e);
		if (!strcmpi(e[0],"D"))
		{
			double l=atof(e[1]);
			sprintf(str,"		x+=%.9lg*xp; y+=%.9lg*yp;\n",l,l);
			strconcat(&ret,str);
		}
		else if (!strcmpi(e[0],"K"))
		{
			double *p=array_1d(m); // [0] not used
			for (i=m-1;i>0;i--) p[i]=atof(e[i])/factorial(i);
			if (!userecurrence)
			{
				for (int ax=0;ax<2;ax++)
				{
					if (ax==0) strconcat(&ret,"		xp+=");
					else strconcat(&ret,"; yp+=");
					char some=0;
					for (i=1;i<m;i++) if (p[i]!=0)
					{
						some=1;
						sprintf(str,"%+.9lg*(%s)",p[i],mpterm(ax,i));
						strconcat(&ret,str);
					}
					if (!some) strconcat(&ret,"0");
				}
				strconcat(&ret,";\n");
			}
			else
			{
				for (i=1;i<m;i++)
				{
					if (i==1) strconcat(&ret,"		X=-x; Y=y;\n");
					else strconcat(&ret,"		oX=X; X=x*X+y*Y; Y=x*Y-y*oX;\n");
					if (p[i]!=0)
					{
						sprintf(str,"		xp+=%+.9lg*X; yp+=%+.9lg*Y;\n",p[i],p[i]);
						strconcat(&ret,str);
					}
				}
			}
			array_1d_free(p);
		}
	}
	strconcat(&ret,"	}\n");
	for (n=0;n<p.iters.m;n++)
	{
		sprintf(str,"	%sif (i<=%d) gl_FragColor=vec4(%g,%g,%g,1);\n",
			n?"else ":"",it[n],
			(float)((n+1)/p.colbase/p.colbase)/(p.colbase-1),(float)((n+1)/p.colbase%p.colbase)/(p.colbase-1),(float)((n+1)%p.colbase)/(p.colbase-1));
		strconcat(&ret,str);
	}
	strconcat(&ret,"	else gl_FragColor=vec4(0,0,0,1);\n\
}");
	return ret;
}

double Problem_domainvolume(const Problem p)
{
	return (p.x1-p.x0)*(p.xp1-p.xp0)*(p.y1-p.y0)*(p.yp1-p.yp0);
}

double Problem_pixelvolume(const Problem p)
{
	return Problem_domainvolume(p)/sq(sq(p.baseres));
}

GLuint fragshaderprog(const char *code)
{
	GLuint fragsh=glCreateShader(GL_FRAGMENT_SHADER);
	glShaderSource(fragsh,1,&code,NULL);
	glCompileShader(fragsh);
	GLuint ret=glCreateProgram();
	glAttachShader(ret,fragsh);
	glLinkProgram(ret);
	return ret;
}

void genquad2(int n,int *pa,int *pb,int *pc,int *pd)
{
	*pa=0; *pb=0; *pc=0; *pd=0;
	for (int i=1;n;i*=2,n/=16)
	{
		if (n&1) {*pa^=i; *pb^=i; *pc^=i; *pd^=i;}
		if (n&2) {*pa^=i; *pb^=i;}
		if (n&4) {*pa^=i; *pc^=i;}
		if (n&8) *pa^=i;
	}
}

double distribute(int n)
{ // 0->0.5, 1->0, 2->0.75, 3->0.25, 4->0.625 etc.
	double ret=0.5*(~n&1),a=0.25;
	for (n>>=1;n;n>>=1,a*=0.5) ret+=a*(n&1);
	return ret;
}

//#define PROFILE
#include <profile1.c>

WINMAIN
{
	int n,i; char str[200];
/*{FILE *out=fopen("genquad2.txt","wt");
for (n=0;n<=2500;n++) {int sa,sb,sc,sd; genquad2(n,&sa,&sb,&sc,&sd);
	fprintf(out,"%d -> %d %d %d %d -> (%lg,%lg,%lg,%lg)\n",n,sa,sb,sc,sd,distribute(sa),distribute(sb),distribute(sc),distribute(sd));}
fclose(out); exit(123);}*/
	gl_windowed=1;
	if (gl_windowed)
	{
		GRAPHICS(1280,1024,"GPU/'fragment shader' programming test [Stephen Brooks 2011]");
	}
	else
	{
		GRAPHICS(0,0,"GPU/'fragment shader' programming test [Stephen Brooks 2011]");
	}
	loadfont("font.dat","normal"); bfmakecopy("normal","profile");
	initshaderprocs();
	Problem p=Problem_load("input.txt");
	char *code=Problem_code(p);
{FILE *out=fopen("shadercode.txt","wt"); fputs(code,out); fclose(out);}
	GLuint prog=fragshaderprog(code);
	FILE *out=fopen("output.csv","wt");
	fprintf(out,"Aperture4D job started %s\n",sjbdatetime());
	fclose(out);
	int rr=p.baseres*p.baseres; unsigned char *d=malloc(rr*rr*3);
	int *c=malloc((1+p.iters.m)*sizeof(int)),*it=p.iters.a;
	double *cs=array_1d_zeroed(1+p.iters.m),*ccs=array_1d_zeroed(1+p.iters.m); int s=0,sm=1;
	double ot,t=eithertime();
	LOOP
	{
		ot=t; t=eithertime();
		GLuint cp;
		/*if (KEY(VK_C))
		{
			int x,y; float isc=1.0/rr,jx=frnd(1),jxp=frnd(1),jy=frnd(1),jyp=frnd(1);
			unsigned c;
			glBegin(GL_POINTS);
			for (x=rr-1;x>=0;x--) for (y=rr-1;y>=0;y--)
			{
				c=cpuver(isc*x,isc*y,jx,jxp,jy,jyp);
				glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
				glVertex2i(x,y);
			}
			glEnd();
		}*/
		profile("glUseProgram");
		glUseProgram(cp=prog);
		profile("genquad2");
		s++;
		int sa,sb,sc,sd; genquad2(s,&sa,&sb,&sc,&sd);
		profile("glUniform1f");
		glUniform1f(glGetUniformLocation(prog,"jx"),distribute(sa));
		glUniform1f(glGetUniformLocation(prog,"jxp"),distribute(sb));
		glUniform1f(glGetUniformLocation(prog,"jy"),distribute(sc));
		glUniform1f(glGetUniformLocation(prog,"jyp"),distribute(sd));
		profile("GL_QUADS");
		glColor3f(0,0,0.5);
		glBegin(GL_QUADS); // This transmits middles of pixels to shader
		glTexCoord2f(0,0); glVertex2i(0,0);
		glTexCoord2f(1,0); glVertex2i(rr,0);
		glTexCoord2f(1,1); glVertex2i(rr,rr);
		glTexCoord2f(0,1); glVertex2i(0,rr);
		glEnd();
		profilec("glFinish",0xFF00FF);
		glFinish();
		profile("reset");
		glUseProgram(0);
		for (n=p.iters.m;n>=0;n--) c[n]=0;
		profile("glReadPixels");
		glReadPixels(0,yres-rr,rr,rr,GL_RGB,GL_UNSIGNED_BYTE,d);
		profile("Counting pixels");
		for (n=rr*rr*3-3;n>=0;n-=3)
		{
			i=((d[n]*(p.colbase-1)+128)/255*p.colbase+(d[n+1]*(p.colbase-1)+128)/255)*p.colbase+(d[n+2]*(p.colbase-1)+128)/255;
			c[i]++;
			/*if (i==0)
			{

				xmin=min(xmin,
			}*/
		}
		c[p.iters.m]+=c[0]; for (n=p.iters.m-1;n>=1;n--) c[n]+=c[n+1];
		double v4scale=Problem_pixelvolume(p),v;
		for (n=0;n<=p.iters.m;n++)
		{
			v=v4scale*c[(n+1)%(p.iters.m+1)];
			cs[n]+=v; ccs[n]+=v*v;
		}
		if (s>=sm*sm*sm*sm)
		{
			out=fopen("output.csv","at");
			fprintf(out,"%d^4 %sgrid\nIters,4D_Vol",p.baseres*sm,(sm&(sm-1)?"irregular ":""));
			if (s>1) fprintf(out,",+/-");
			fprintf(out,"\n");
			for (n=0;n<=p.iters.m;n++)
			{
				if (n>0) fprintf(out,"%d",it[n-1]);
				else fprintf(out,"Domain");
				fprintf(out,",%.9lg",cs[n]/s);
				if (s>1) fprintf(out,",%.9lg",sqrt(noneg(ccs[n]-sq(cs[n])/s)/((s-1)*s)));
				fprintf(out,"\n");
			}
			fprintf(out,"\n");
			fclose(out);
			sm++;
		}
		profile_display(yres*2/3,0,1,0);
		sprintf(str,"prog=%d glGetError()=%d (GL_NO_ERROR=%d)\n%lg samples/s\n%lg particles/s",
			cp,glGetError(),GL_NO_ERROR,1.0/(t-ot),(double)rr*rr/(t-ot));
		bfwrite(yres,0,str,GFX_white);
		sprintf(str,"Sample %d (%d^4 in %d)\n%lg particles",
			s,sm,sm*sm*sm*sm-s,(double)rr*rr*s);
		bfwrite(yres,BF_height*5,str,GFX_white);
		if (KHIT(VK_S)) screenshotauto();
		ccl();
	}
}

